#include "acquire_control.h"
#include "bus/event_queue.h"
#include "bus/event_handler.h"
#include "com/debug.h"
#include "gui/infoled.h"

#include "malloc_safe.h"
#include "sampling.h"
#include "globals.h"
#include <math.h>
#include "gui/wifistatusled.h"

// ----

static volatile bool acq_running = false;

static struct {
	void (*sucCb)(uint32_t*, uint16_t, void*);
	void (*failCb)(void*);
	void* xtra;
} cbinfo;

static task_pid_t abort_tmeo_pid;

static uint32_t h1, h2;

// for wdg
bool acq_is_running(void)
{
	return acq_running;
}


static void indicate_capture(bool capturing)
{
	// TODO use second LED.

	if (capturing) {
		// red LED will indicate sampling
		set_acqstatus_led(ACQLED_CAPTURE_COLOR);
	} else {
		set_acqstatus_led(ACQLED_IDLE_COLOR);
	}
}


static bool acquire_task_done(uint32_t handler_id, Event *evt, void **user_data)
{
	UNUSED_ARG(user_data);

	uint32_t *buf = (uint32_t*)evt->data;

	acq_running = false;
	indicate_capture(false);
	abort_scheduled_task(abort_tmeo_pid);

	if (evt->type == EVT_ACQ_DMA_COMPLETE) {
		// evt->data_ptr ... buffer
		// evt->num_a ...... buffer length
		// evt->num_b ...... buffer width

		dbg("DMA completed, buffer = %"PRIu32"x%"PRIu32, evt->num_a, evt->num_b);

		// the callback is responsible for freeing the buffer and xtra
		cbinfo.sucCb(buf, (uint16_t)evt->num_a, cbinfo.xtra);
	} else {
		// fail
		if (cbinfo.failCb) {
			cbinfo.failCb(cbinfo.xtra);
		} else {
			warn("Fail CB missing in capture_task.");
			// WARN if this was not malloc'd, we may get an error.
			free(cbinfo.xtra); // (if NULL, no-op)
		}
	}

	// This is reached in all cases:

	// removes also the chained handler
	remove_event_handler(handler_id);
	SoftWDG_Feed(SoftWDG_ACQ_IDLE);

	return true; // event was handled.
}

// timeout task
static void abort_acq(void *unused)
{
	(void)unused;

	error("Sampling timed out - sending FAIL event");
	remove_event_handler(h1);

	free(cbinfo.xtra);

	Event evt;
	evt.type = EVT_ACQ_DMA_FAILED;
	evt.data = NULL;
	eq_post(&evt);
}

bool start_capture_task(
	uint32_t *buf,
	uint16_t count,
	uint32_t frequency,
	void (*sucCb)(uint32_t* buf, uint16_t, void*),
	void (*failCb)(void*),
	void *xtra)
{
	if (acq_running) {
		warn("ACQ is in progress.");
		return false;
	}

	indicate_capture(true);

	// info to be passed to the handlers
	cbinfo.sucCb = sucCb;
	cbinfo.failCb = failCb;
	cbinfo.xtra = xtra;

	// set handler
	h1 = register_event_handler(EVT_ACQ_DMA_COMPLETE, acquire_task_done, NULL); // we could use the xtra / user_data, but no need here
	h2 = register_event_handler(EVT_ACQ_DMA_FAILED, acquire_task_done, NULL);   //  static var works fine.
	if (!h1 || !h2) {
		// either of them failed, remove both and abandon the attempt
		remove_event_handler(h1);
		remove_event_handler(h2);
		goto fail;
	}
	chain_event_handler(h1, h2, true); // chain with each other

	// set timeout
	ms_time_t estimate = (uint32_t)lroundf((count * 1000.0f) / frequency + 200);
	dbg("Timeout = %"PRIu32" ms", estimate);
	abort_tmeo_pid = schedule_task(abort_acq, NULL, estimate, true);

	// start capture
	bool suc = acq_start_dma_capture(buf, count, frequency, DMADataWidth_32);

	// if failed to start, clean up.
	if (suc) {
		acq_running = true;
	} else {
		goto fail;
	}

	return true;

fail:
	abort_scheduled_task(abort_tmeo_pid);
	indicate_capture(false);
	return false;
}
