#include "hw_init.h"

#include "hw_utils/all.h"

#include "com/iface_usart.h"
#include "com/com_fileio.h"
#include "com/datalink.h"

#include "utils/debounce.h"
#include "utils/timebase.h"

#include "gui/buttons.h"
#include "gui/wifistatusled.h"

#include "bus/event_queue.h"

#include "acquire/sampling.h"

// ---- Private prototypes --------

static void conf_gpio(void);
static void conf_usart(void);
static void conf_systick(void);
static void conf_subsystems(void);
static void conf_watchdog(void);
static void conf_irq_prios(void);


// ---- Public functions ----------

/**
 * @brief Initialize hardware resources
 */
void hw_init(void)
{
	conf_gpio();
	conf_usart();
	conf_systick();
	conf_watchdog();

	conf_irq_prios();

	// initialize HW for ADC
	acq_init();

	conf_subsystems();
}


// ---- Private functions ---------


static void conf_watchdog(void)
{
	WDG_Configure(5000);
}


static void conf_irq_prios(void)
{
	NVIC_SetPriorityGrouping(0); // 0 bits for sub-priority

	// SysTick - highest prio, used for timeouts
	NVIC_SetPriority(SysTick_IRQn, 0); // SysTick - for timeouts
	NVIC_SetPriority(DMA1_Channel1_IRQn, 5); // DMA1_ch1 - ACQ
	NVIC_SetPriority(USART2_IRQn, 6); // USART - datalink
	NVIC_SetPriority(USART1_IRQn, 10); // USART - debug
}


/**
 * @brief Configure SW subsystems
 */
static void conf_subsystems(void)
{
	// task scheduler subsystem
	size_t peri_cnt = 20;
	size_t fut_cnt = 20;
	timebase_init(peri_cnt, fut_cnt);

	// event and task queues
	size_t tq_len = 100;
	size_t eq_len = 100;
	queues_init(tq_len, eq_len);

	// debouncer task @ 1 ms
	add_periodic_task(debo_periodic_task, NULL, 1, false);

	// set up debouncing for buttons
	buttons_init();

	// initialize SBMP for ESP8266
	dlnk_init();

	// wifi colors
	wifistatusled_init();
}


/**
 * @brief Configure GPIOs
 */
static void conf_gpio(void)
{
	// Enable clocks
	RCC->AHBENR |= (RCC_AHBENR_GPIOAEN | RCC_AHBENR_GPIOBEN | RCC_AHBENR_GPIOCEN);

	gpio_conf_input(GPIOC, GPIO_PuPd_UP, GPIO13); // AP
	gpio_conf_input(GPIOC, GPIO_PuPd_UP, GPIO14); // WPS

	gpio_conf_output(GPIOC, GPIO_Speed_10MHz, GPIO_OType_PP, GPIO15); // RGB data line
	gpio_conf_output(GPIOA, GPIO_Speed_10MHz, GPIO_OType_PP, GPIO4); // Status LED

	// Debug USART
	gpio_conf_af(GPIOB, GPIO_Speed_10MHz, GPIO_OType_PP, GPIO_AF_7, GPIO6 | GPIO7); // debug

	// ESP8266
	gpio_conf_af(GPIOA, GPIO_Speed_10MHz, GPIO_OType_PP, GPIO_AF_7, GPIO2 | GPIO3); // ESP rx, tx
	gpio_conf_output(GPIOA, GPIO_Speed_2MHz, GPIO_OType_PP, GPIO1); // ESP reset

	// ADC input
	gpio_conf_analog(GPIOA, GPIO0);
}


/**
 * @brief Configure USARTs
 */
static void conf_usart(void)
{
	// Debug interface, working as stdout/stderr. Stdin is not implemented.
	debug_iface = usart_iface_init(USART1, 115200, 256, 256);
	setvbuf(stdout, NULL, _IONBF, 0);
	setvbuf(stderr, NULL, _IONBF, 0);
	setvbuf(stdin, NULL, _IONBF, 0);
	debug_iface->file = stdout;


	// Datalink iface
	data_iface = usart_iface_init(USART2, 460800, 256, 256);
//	FILE *x = fopen(FNAME_DLNK, "r+"); // read/write - using com_fileio
//	setvbuf(x, NULL, _IONBF, 0);
//	data_iface->file = x;
}


/**
 * @brief Configure 1 kHz SysTick w/ interrupt
 */
static void conf_systick(void)
{
	SysTick_Config(F_CPU / 1000);
}
